import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart' as plugin;
import 'package:analyzer_plugin/protocol/protocol_generated.dart' as plugin;

class LintError {
  const LintError({
    required this.message,
    required this.code,
    required this.node,
  });

  static final String _exhaustiveKeysCode = 'exhaustive_keys';
  static final String _nestedHooks = 'nested_hooks';

  factory LintError.missingKey(String key, String? kind, AstNode node) {
    return LintError(
      message: "Missing key '$key' " +
          (kind != null ? '($kind) ' : '') +
          'found. Add the key, or ignore this line. ',
      code: _exhaustiveKeysCode,
      node: node,
    );
  }

  factory LintError.unnecessaryKey(String key, String? kind, AstNode node) {
    return LintError(
      message: "Unnecessary key '$key' " +
          (kind != null ? '($kind) ' : '') +
          'found. Remove the key, or ignore this line.',
      code: _exhaustiveKeysCode,
      node: node,
    );
  }

  factory LintError.functionKey(
    String functionName,
    AstNode node,
  ) {
    return LintError(
      message:
          "'$functionName' changes on every re-build. Move its definition inside the hook, or wrap with useCallback.",
      code: _exhaustiveKeysCode,
      node: node,
    );
  }

  factory LintError.nestedHooks(String hookName, AstNode node) {
    return LintError(
      message:
          'Avoid nested use of $hookName. Hooks must be used in top-level scope of the build function.',
      code: _nestedHooks,
      node: node,
    );
  }

  final String message;
  final String code;
  final AstNode node;

  plugin.AnalysisError toAnalysisError(String file, CompilationUnit unit) {
    final location = _toLocation(file, unit);

    return plugin.AnalysisError(
      plugin.AnalysisErrorSeverity.INFO,
      plugin.AnalysisErrorType.LINT,
      location,
      message,
      code,
    );
  }

  plugin.Location _toLocation(String file, CompilationUnit unit) {
    final lineInfo = unit.lineInfo!;
    final begin = node.beginToken.charOffset;
    final end = node.endToken.charEnd;
    final loc = lineInfo.getLocation(begin);
    final locEnd = lineInfo.getLocation(end);

    return plugin.Location(
      file,
      node.beginToken.charOffset,
      node.length,
      loc.lineNumber,
      loc.columnNumber,
      endLine: locEnd.lineNumber,
      endColumn: locEnd.columnNumber,
    );
  }
}

class LintFix {
  const LintFix({
    required this.message,
    required this.node,
    required this.replacement,
  });

  final String message;
  final AstNode node;
  final String replacement;

  plugin.PrioritizedSourceChange toAnalysisFix(
    String file,
    ResolvedUnitResult analysisResult,
  ) {
    return plugin.PrioritizedSourceChange(
      1,
      plugin.SourceChange(
        message,
        edits: [
          plugin.SourceFileEdit(
            file,
            analysisResult.libraryElement.source.modificationStamp,
            edits: [
              plugin.SourceEdit(
                node.beginToken.charOffset,
                node.length,
                replacement,
              ),
            ],
          ),
        ],
      ),
    );
  }
}
